#include <string.h>

#include "model_int2.h"

#define DEBUG 0

void (*set_bnd)(int b, freal *a);
void model_set_bnd_rout(void *func)
{
	printf("bnd_rout (before) =0x%08X\n", set_bnd);
	printf("void ptr argument=0x%08X\n", func);
	set_bnd=func;
	printf("bnd_rout (after) =0x%08X\n", set_bnd);
}
//void set_bnd(int b, freal *a) {}

/* apply external density influence */
void add_source(freal *x, freal *s, freal dt)
{
	int i,j;

	#if 0
	//dt=X2I(dt<<10);
	//for (i=SIZE-1; i>=0; --i) x[i]+=(dt*s[i])>>10;
	for (i=SIZE-1; i>=0; --i) x[i]+=s[i]>>6;
	#else
	for (i=SIZE-1; i>=0; --i) x[i]+= XM(dt,s[i]);
	#endif

	#if DEBUG
	printf("add_source()\n");
	for (j=0; j<=HEIGHT+1; j++) {
		printf("s = ");
		for (i=0; i<=WIDTH+1; i++) {
			printf("%04X ", s[IX(i,j)]);
		}
		printf("\n");
	}
	printf("\n");

	for (j=0; j<=HEIGHT+1; j++) {
		printf("x = ");
		for (i=0; i<=WIDTH+1; i++) {
			printf("%04X ", x[IX(i,j)]);
		}
		printf("\n");
	}
	printf("\n");
	#endif
}

void diffuse_bad(int b, freal *x, freal *x0, freal diff, freal dt)
{
	int i,j;
	freal a=XM(dt,diff)*WIDTH*HEIGHT;

	for (i=BORDER; i<WIDTH+BORDER; i++) {
		for (j=BORDER; j<HEIGHT+BORDER; j++) {
			x[IX(i,j)] = x0[IX(i,j)]
			           + XM(a, ( x0[IX(i-1,j)]+x0[IX(i+1,j)]+x0[IX(i,j-1)]+x0[IX(i,j+1)]-4*x0[IX(i,j)] ));
		}
	}
	set_bnd(b, x);
}

/* gauss seidel stabilised.. still grows too much in some cases :/ */
void diffuse(int b, freal *x, freal *x0, freal diff, freal dt)
{
	int i, j, k;
	freal a=XM(dt*WIDTH,diff)*HEIGHT;

	for (k=0; k<20; k++) {
		for (j=BORDER; j<HEIGHT+BORDER; j++) {
			for (i=BORDER; i<WIDTH+BORDER; i++) {
				x[IX(i,j)] = (
			                       X1_0 * x0[IX(i,j)]
				             + a * (x[IX(i-1,j)]+x[IX(i+1,j)]+x[IX(i,j-1)]+x[IX(i,j+1)])
					     )
					     / (X1_0+4*a);
			}
		}
		set_bnd(b, x);
	}
}

void diffuse_fixed(int b, freal *x, freal *x0, freal diff, freal dt)
{
	int i, j, k;
	freal a=XM(dt*WIDTH,diff)*HEIGHT;
	long s1 = ((long long) X1_0<<12)/((X1_0+4*a));
	long s2 = ((long long) a<<12)/((X1_0+4*a));

	for (j=BORDER; j<HEIGHT+BORDER; j++) {
		for (i=BORDER; i<WIDTH+BORDER; ++i) {
			#if FIXEDPOINT
			x[IX(i,j)] =
			(
			  s1*(x0[IX(i,j)]>>4) + s2*((x0[IX(i-1,j)]+x0[IX(i+1,j)]+x0[IX(i,j-1)]+x0[IX(i,j+1)])>>4)
			)>>8;
			#else
			x[IX(i,j)] = (
				       XML( X1_0, x0[IX(i,j)] )
				     + XML( a, (x0[IX(i-1,j)]+x0[IX(i+1,j)]+x0[IX(i,j-1)]+x0[IX(i,j+1)]) )
				     )
				     / (X1_0+4*a);
			#endif
		}
	}
	set_bnd(b, x);

	#if DEBUG
	printf("diffuse_fixed(dt=0x%08X * diff=0x%08X = a=0x%08X)\n", dt, diff, a);
	for (j=0; j<=HEIGHT+1; j++) {
		printf("x0 = ");
		for (i=0; i<=WIDTH+1; i++) {
			printf("%04X ", x0[IX(i,j)]);
		}
		printf("\n");
	}
	printf("\n");

	for (j=0; j<=HEIGHT+1; j++) {
		printf("x = ");
		for (i=0; i<=WIDTH+1; i++) {
			printf("%04X ", x[IX(i,j)]);
		}
		printf("\n");
	}
	printf("\n");
	#endif
}

void advect(int b, freal *d, freal *d0, freal *u, freal *v, freal dt)
{
	int i,j,i0,j0,i1,j1;
	freal x,y,s0,t0,s1,t1,dt0;

	//printf("d=0x%08X d0=0x%08X u=0x%08X v=0x%08X\n", d, d0, u, v);

	dt0=HALVE( (dt * (WIDTH+HEIGHT)) );
	//printf("dt0=0x%X\n", dt0);
	for (j=BORDER; j<HEIGHT+BORDER; j++) {
		//printf("i=%d\n", i);

		for (i=BORDER; i<WIDTH+BORDER; i++) {
			/* backtrack to previous grid position.. */
			//printf("j=%d\t", j);
			x = I2X(i) - XM(dt0, u[IX(i,j)]);
			y = I2X(j) - XM(dt0, v[IX(i,j)]);

			unsigned long xs = (unsigned long) (x-HALVE(X1_0));
			if (xs > I2X(WIDTH)) {
				if ((long) xs < 0)
					x = HALVE(X1_0);
				else if (xs > I2X(WIDTH))
					x=I2X(WIDTH)+HALVE(X1_0);
			}
			i0=X2I(x);
			i1=i0+1;

			#if 1
			unsigned long ys = (unsigned long) (y-HALVE(X1_0));
			if (ys > I2X(WIDTH)) {
				if ((long) ys < 0)
					y = HALVE(X1_0);
				else if (ys > I2X(HEIGHT))
					y=I2X(HEIGHT)+HALVE(X1_0);
			}
			#else
			if (y < HALVE(X1_0))
				y = HALVE(X1_0);
			else if ( y > I2X(HEIGHT)+HALVE(X1_0) )
				y = I2X(HEIGHT)+HALVE(X1_0);
			#endif
			j0=X2I(y);
			j1=j0+1;

			s1=x-I2X(i0);
			s0=X1_0-s1;
			//printf("s0=0x%X ", s0);

			t1=y-I2X(j0);
			t0=X1_0-t1;
			//printf("t0=0x%X ", t0);

			/* bilinear interpolation.. */
			d[IX(i,j)] = XM(s0, (XM(t0, d0[IX(i0,j0)]) + XM(t1, d0[IX(i0,j1)])) )
			           + XM(s1, (XM(t0, d0[IX(i1,j0)]) + XM(t1, d0[IX(i1,j1)])) );
		}
	}
	//set_bnd(b, d);
}

/* bilinear interpolation.. */
#define INTERP() \
i0=X2I(x); \
i1=i0+1; \
j0=X2I(y); \
j1=j0+1; \
s1=(x-I2X(i0))>>11; \
s0=0x100-s1; \
t1=(y-I2X(j0))>>11; \
t0=0x100-t1; \
d[IX(i,j)] = \
( \
  s0 * ((t0 * d0[IX(i0,j0)] + t1 * d0[IX(i0,j1)])>>8 ) \
+ s1 * ((t0 * d0[IX(i1,j0)] + t1 * d0[IX(i1,j1)])>>8 ) \
)>>8;
//i0=X2I(x); \
//i1=i0+1; \
//j0=X2I(y); \
//j1=j0+1; \
//s1=x-I2X(i0); \
//s0=X1_0-s1; \
//t1=y-I2X(j0); \
//t0=X1_0-t1; \
//d[IX(i,j)] = \
//  XM(s0, (XM(t0, d0[IX(i0,j0)]) + XM(t1, d0[IX(i0,j1)])) ) \
//+ XM(s1, (XM(t0, d0[IX(i1,j0)]) + XM(t1, d0[IX(i1,j1)])) );

#define BACKTRACK() \
x = I2X(i) - dt0 * u[IX(i,j)]; \
y = I2X(j) - dt0 * v[IX(i,j)];

/*
 * optimised version with respect to clipping.
 * clipping is done only at the side pixels..
 * under the assumption that a cell cannot move more than one unit per frame..
 * this is reasonable because this is also a criterion for every boundary condition.
 * this is significantly faster..
 */
void advect_new(int b, freal *d, freal *d0, freal *u, freal *v, freal dt)
{
	int i,j,i0,j0,i1,j1;
	freal x,y,s0,t0,s1,t1,dt0;
	//printf("d=0x%08X d0=0x%08X u=0x%08X v=0x%08X\n", d, d0, u, v);
	/* dt0 is big enough and can be respresented as an int.. */
	dt0=X2I(HALVE( dt * (WIDTH+HEIGHT) ));

	/* top case.. check for out of bounds extrapolation.. */
	i=BORDER;
	j=BORDER;
	BACKTRACK();
	if (x < HALVE(X1_0))
		x = HALVE(X1_0);
	if (y < HALVE(X1_0))
		y = HALVE(X1_0);
	INTERP();
	for (i=BORDER+1; i<WIDTH+BORDER-1; i++) {
		/* backtrack to previous grid position.. */
		BACKTRACK();
		if (y < HALVE(X1_0))
			y = HALVE(X1_0);
		INTERP();
	}
	BACKTRACK();
	if (x > I2X(WIDTH)+HALVE(X1_0))
		x=I2X(WIDTH)+HALVE(X1_0);
	if (y < HALVE(X1_0))
		y = HALVE(X1_0);
	INTERP();

	/* body case.. don't check for out of bounds extrapolation in Y dir.. */
	for (j=BORDER+1; j<HEIGHT+BORDER-1; j++) {
		i=BORDER;
		BACKTRACK();
		if (x < HALVE(X1_0))
			x = HALVE(X1_0);
		INTERP();

		for (i=BORDER+1; i<WIDTH+BORDER-1; i++) {
			/* backtrack to previous grid position.. */
			BACKTRACK();
			INTERP();
		}

		BACKTRACK();
		if (x > I2X(WIDTH)+HALVE(X1_0))
			x=I2X(WIDTH)+HALVE(X1_0);
		INTERP();
	}

	/* bottom case.. check for out of bounds extrapolation.. */
	i=BORDER;
	BACKTRACK();
	if (x < HALVE(X1_0))
		x = HALVE(X1_0);
	if (y > I2X(HEIGHT)+HALVE(X1_0))
		y = I2X(HEIGHT)+HALVE(X1_0);
	INTERP();
	for (i=BORDER+1; i<WIDTH+BORDER-1; i++) {
		/* backtrack to previous grid position.. */
		BACKTRACK();
		if (y > I2X(HEIGHT)+HALVE(X1_0))
			y = I2X(HEIGHT)+HALVE(X1_0);
		INTERP();
	}
	BACKTRACK();
	if (x > I2X(WIDTH)+HALVE(X1_0))
		x=I2X(WIDTH)+HALVE(X1_0);
	if (y > I2X(HEIGHT)+HALVE(X1_0))
		y = I2X(HEIGHT)+HALVE(X1_0);
	INTERP();

	//set_bnd(b, d);
}

freal tmpbuf[SIZE];

/* i just don't get the add source in this case.. it only works when x0 (prev_dens) is cleared except for new sources */
void dens_step(freal *x, freal *x0, freal *u, freal *v, freal diff, freal dt)
{
	advect_new(0,x,x0,u,v,dt);
}

#define GS_NUM_LOOPS 20

__inline void diag(int i, freal *p, freal *div)
{
	int j;
	int end;
	if (i<WIDTH) {
		end = (i>=HEIGHT) ? end=i-HEIGHT+1 : 0;
		for (j=i; j>=end; --j) {
			int x = BORDER+j;
			int y = BORDER+i-j;
			p[IX(x,y)] =
			 (
			   div[IX(x,y)]
			 + p[IX(x-1,y)]+p[IX(x+1,y)]+p[IX(x,y-1)]+p[IX(x,y+1)]
			 )/4;
		}
	} else if (i<WIDTH+HEIGHT) {
		int y_off = i-WIDTH;
		for (j=WIDTH+HEIGHT-i-1; j>0; --j) {
			int x = BORDER+WIDTH-j;
			int y = BORDER+y_off+j;
			p[IX(x,y)] =
			 (
			   div[IX(x,y)]
			 + p[IX(x-1,y)]+p[IX(x+1,y)]+p[IX(x,y-1)]+p[IX(x,y+1)]
			 )/4;
		}
	}
}

freal p[SIZE], div[SIZE];

void project(freal *u, freal *v, int gs_loops)
{
	int i,j,k,m;
	freal h;

	#if 1
	for (i=0; i<WIDTH+BORDER*2; ++i) {
		p[IX(i,0)]=0;
		p[IX(i,HEIGHT+BORDER)]=0;
		div[IX(i,0)]=0;
		div[IX(i,HEIGHT+BORDER)]=0;
	}
	for (j=0; j<HEIGHT+BORDER*2; ++j) {
		p[IX(0,j)]=0;
		p[IX(WIDTH+BORDER,j)]=0;
		div[IX(0,j)]=0;
		div[IX(WIDTH+BORDER,j)]=0;
	}
	#endif

	h = I2X(2)/(HEIGHT+WIDTH);
	long sc = -h/2;
	for (j=BORDER; j<HEIGHT+BORDER; j++) {
		for (i=BORDER; i<WIDTH+BORDER; i++) {
			div[IX(i,j)] = XM( sc, (u[IX(i+1,j)]-u[IX(i-1,j)]+v[IX(i,j+1)]-v[IX(i,j-1)]) );
			p[IX(i,j)]=0;
		}
	}
	set_bnd(0,div);
	set_bnd(0,p);

	#if 0

	/* optimised for better data cache behaviour, there should be more hits */
	/* after some tests it turned out this is not faster at all :( */
	for (m=0; m<WIDTH+HEIGHT+gs_loops; ++m) {
		int end=m;
		int start=m-gs_loops;
		if (start<0) {
			start=0;
		}
		if (end>=WIDTH+HEIGHT) {
			end=WIDTH+HEIGHT-1;
		}
		for (k=end; k>m-gs_loops; --k) {
			diag(k, p, div);
			//print_array();
		}
	}

	#else

	for (k=0; k<gs_loops; k++) {
		for (j=BORDER; j<HEIGHT+BORDER; j++) {
			int y_offset=j*(WIDTH+BORDER*2);
			#ifdef GP2X
			/* you don't have to believe me but this asm version is much better than the one from gcc 3.4.6 -O2 */
			/* %0: div[], %1: p[], %2: y_offset */
			__asm__ __volatile__(
			"mov     r6, #160\n\t"
			".loop:\n\t"
			"add     r0, %2, r6\n\t"
			"add     r1, %1, r0, asl #2\n\t"
			"ldr     r3, [%0, r0, asl #2]\n\t"
			"ldr     r2, [r1, #-4]\n\t"
			"ldr     ip, [r1, #4]\n\t"
			"ldr     r7, [r1, #-648]\n\t"
			"add     r3, r3, r2\n\t"
			"ldr     r2, [r1, #648]\n\t"
			"add     r3, r3, ip\n\t"
			"add     r3, r3, r7\n\t"
			"add     r3, r3, r2\n\t"
			"subs    r6, r6, #1\n\t"
			"mov     r3, r3, asr #2\n\t"
			"str     r3, [%1, r0, asl #2]\n\t"
			"bne     .loop\n\t"
			:
			: "r" (div), "r" (p), "r" (y_offset)
			: "cc", "memory", "r0", "r1", "r2", "r3", "r6", "r7", "ip"
			);
			#else
			for (i=BORDER; i<WIDTH+BORDER; i++) {
				p[IX(i,j)] = (
				               div[IX(i,j)]
					     + p[IX(i-1,j)]+p[IX(i+1,j)]+p[IX(i,j-1)]+p[IX(i,j+1)]
				             )/4;
			}
			#endif
		}
		set_bnd(0,p);
	}

	#endif

	freal inv = XDL(X1_0, 2*h);
	long s = (WIDTH+HEIGHT)>>2;
	for (j=BORDER; j<HEIGHT+BORDER; j++) {
		for (i=BORDER; i<WIDTH+BORDER; i++) {
			//u[IX(i,j)] -= XM(p[IX(i+1,j)]-p[IX(i-1,j)], inv);
			//v[IX(i,j)] -= XM(p[IX(i,j+1)]-p[IX(i,j-1)], inv);
			u[IX(i,j)] -= s*(p[IX(i+1,j)]-p[IX(i-1,j)]);
			v[IX(i,j)] -= s*(p[IX(i,j+1)]-p[IX(i,j-1)]);
		}
	}
	set_bnd(1,u);
	set_bnd(2,v);
}

freal tmpu[SIZE], tmpv[SIZE];

void vel_step(freal *u, freal *v, freal visc, freal dt)
{
	#if DEBUG
	printf("vel_step(visc=0x%08X)\n", visc);
	#endif

	diffuse_fixed(1,tmpu,u,visc,dt);
	diffuse_fixed(2,tmpv,v,visc,dt);
	project(tmpu,tmpv,5);
	advect_new(1,u,tmpu,tmpu,tmpv,dt);
	advect_new(2,v,tmpv,tmpu,tmpv,dt);
	project(u,v,20);
}
