#include <SDL/SDL.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

#include "model_int2.h"

#define OUTPUT_IMAGES 0
#define SINGLE_PGM 0

/* velocity maps (u,v) t,t-1 */
static freal u[SIZE], v[SIZE], u_prev[SIZE], v_prev[SIZE];
/* density maps t,t-1 */
static freal dens[SIZE], dens_prev[SIZE];
static freal *log_dens_ptr, *phys_dens_ptr;

FILE *fout;

#define SDL_WIDTH 320
#define SDL_HEIGHT 240

#define GP2X_BUTTON_UP              (0)
#define GP2X_BUTTON_DOWN            (4)
#define GP2X_BUTTON_LEFT            (2)
#define GP2X_BUTTON_RIGHT           (6)
#define GP2X_BUTTON_UPLEFT          (1)
#define GP2X_BUTTON_UPRIGHT         (7)
#define GP2X_BUTTON_DOWNLEFT        (3)
#define GP2X_BUTTON_DOWNRIGHT       (5)
#define GP2X_BUTTON_CLICK           (18)
#define GP2X_BUTTON_A               (12)
#define GP2X_BUTTON_B               (13)
#define GP2X_BUTTON_X               (14)
#define GP2X_BUTTON_Y               (15)
#define GP2X_BUTTON_L               (10)
#define GP2X_BUTTON_R               (11)
#define GP2X_BUTTON_START           (8)
#define GP2X_BUTTON_SELECT          (9)
#define GP2X_BUTTON_VOLUP           (16)
#define GP2X_BUTTON_VOLDOWN         (17)

SDL_Surface *screen;
SDL_Color colors[256];
int sound;

/*---------------------------------------------------------------------------*/
/* boundary condition code                                                   */
/*---------------------------------------------------------------------------*/

#define BALL 1
#define BIG_BALL 1
#if BIG_BALL
# define D 11
#else
# define D 9
#endif
#define D2 (D/2)
unsigned char ball[D*D]=
{
#if BIG_BALL
	0,0,0,0,0,1,0,0,0,0,0,
	0,0,0,1,1,1,1,1,0,0,0,
	0,0,1,1,1,1,1,1,1,0,0,
	0,1,1,1,1,1,1,1,1,1,0,
	0,1,1,1,1,1,1,1,1,1,0,
	1,1,1,1,1,1,1,1,1,1,1,
	0,1,1,1,1,1,1,1,1,1,0,
	0,1,1,1,1,1,1,1,1,1,0,
	0,0,1,1,1,1,1,1,1,0,0,
	0,0,0,1,1,1,1,1,0,0,0,
	0,0,0,0,0,1,0,0,0,0,0
#else
	0,0,0,0,1,0,0,0,0,
	0,0,1,1,1,1,1,0,0,
	0,1,1,1,1,1,1,1,0,
	0,1,1,1,1,1,1,1,0,
	1,1,1,1,1,1,1,1,1,
	0,1,1,1,1,1,1,1,0,
	0,1,1,1,1,1,1,1,0,
	0,0,1,1,1,1,1,0,0,
	0,0,0,0,1,0,0,0,0
#endif
};
unsigned char ball_occupied[D*D];

#define BIX(x,y) ((x)+D2+ D*((y)+D2))

void set_bnd_ball(int b, freal *a)
{
	int i,x,y;

	for (i=BORDER; i<WIDTH+BORDER; i++) {
		a[IX(i, BORDER-1)       ] = b==2 ? -a[IX(i,BORDER)] : a[IX(i,BORDER)];
		//a[IX(i, HEIGHT+BORDER)] = b==2 ? -a[IX(i,HEIGHT+BORDER-1)] : a[IX(i,HEIGHT+BORDER-1)];
	}
	for (i=BORDER; i<HEIGHT+BORDER; i++) {
		/* disable if we want a wind tunnel! */
		//a[IX(BORDER-1   ,i)    ] = b==1 ? -a[IX(BORDER,i)] : a[IX(BORDER,i)];
		//a[IX(WIDTH+BORDER ,i)] = b==1 ? -a[IX(WIDTH+BORDER-1,i)] : a[IX(WIDTH+BORDER-1,i)];
	}

	/* basin corners */
	// todo: take border into account!
	//a[IX(0  ,0  )         ] = .5*(a[IX(1,0  )         ]+a[IX(0  ,1)]);
	//a[IX(0  ,HEIGHT+1)    ] = .5*(a[IX(1,HEIGHT+1)    ]+a[IX(0  ,HEIGHT)]);
	//a[IX(WIDTH+1,0  )     ] = .5*(a[IX(WIDTH,0  )     ]+a[IX(WIDTH+1,1)]);
	//a[IX(WIDTH+1,HEIGHT+1)] = .5*(a[IX(WIDTH,HEIGHT+1)]+a[IX(WIDTH+1,HEIGHT)]);

	#if BALL
	/* an object blocking the stream.. */
	memset(ball_occupied, 0, D*D);
	int xoff=BORDER+(WIDTH/10), yoff=BORDER+(HEIGHT/2);

	/* build left and right walls */
	for (y=-D2; y<=D2; y++) {
		int in=0;
		for(x=-D2; x<=D2; x++) {
			int bool=ball[BIX(x,y)];
			if (bool && !in) {
				/* make a left wall */
				a[IX(x+xoff,y+yoff)]= b&1 ? -a[IX(x+xoff-1,y+yoff)] : a[IX(x+xoff-1,y+yoff)];
				in=1;
				ball_occupied[BIX(x,y)]=1;
			} else if (!bool && in) {
				/* make a right wall */
				a[IX(x+xoff,y+yoff)]= b&1 ? -a[IX(x+xoff+1,y+yoff)] : a[IX(x+xoff+1,y+yoff)];
				in=0;
				ball_occupied[BIX(x,y)]=1;
			}
		}
	}
	/* build top and bottom walls */
	for (x=-D2; x<=D2; x++) {
		int in=0;
		for(y=-D2; y<=D2; y++) {
			int bool=ball[BIX(x,y)];

			if (!ball_occupied[BIX(x,y)]) {
				if (bool && !in) {
					/* make a top wall */
					a[IX(x+xoff,y+yoff)]= b&2 ? -a[IX(x+xoff,y+yoff-1)] : a[IX(x+xoff,y+yoff-1)];
					in=1;
				} else if (!bool && in) {
					/* make a bottom wall */
					a[IX(x+xoff,y+yoff)]= b&2 ? -a[IX(x+xoff,y+yoff+1)] : a[IX(x+xoff,y+yoff+1)];
					in=0;
				}
			}

		}

	}
	#endif /*#if BALL*/

}

void dampen_ball(freal *dens) 
{
	int x,y;
	int xoff=BORDER+(WIDTH/10), yoff=BORDER+(HEIGHT/2);
	for (x=-D2; x<=D2; x++) {
		for(y=-D2; y<=D2; y++) {
			if (ball[BIX(x,y)])
				dens[IX(x+xoff,y+yoff)]=HALVE(dens[IX(x+xoff,y+yoff)]);
		}

	}
	
}


/* control parameters */
static int v_dir=0, gas=0, blow=0;

/* modify dens, u, v buffers depending on incoming key events.. */
int get_from_ui_windtunnel(freal *dens, freal *u, freal *v)
{
	int i,j,x,y;

	SDL_Event event;
	while ( SDL_PollEvent(&event) )	{
		if (event.type == SDL_QUIT) {
			return -1;
		}
		switch(event.type) {
			case SDL_KEYDOWN:
			switch(event.key.keysym.sym) {
				case SDLK_ESCAPE:
				return -1;
				case SDLK_SPACE:
				blow=1;
				break;
				case SDLK_RETURN:
				gas=1;
				break;
				case SDLK_UP:
				v_dir=-1;
				break;
				case SDLK_DOWN:
				v_dir=+1;
				break;
			}
			break;
			case SDL_KEYUP:
			switch(event.key.keysym.sym) {
				case SDLK_SPACE:
				blow=0;
				break;
				case SDLK_RETURN:
				gas=0;
				break;
				case SDLK_UP:
				v_dir=0;
				break;
				case SDLK_DOWN:
				v_dir=0;
				break;
			}
			break;
			case SDL_JOYBUTTONDOWN:
			switch(event.jbutton.button) {
				case GP2X_BUTTON_START:
				return -1;
				case GP2X_BUTTON_X:
				gas=1;
				break;
				case GP2X_BUTTON_UP:
				v_dir=-1;
				break;
				case GP2X_BUTTON_DOWN:
				v_dir=+1;
				break;
				case GP2X_BUTTON_A:
				blow=1;
				break;
			}
			break;
			case SDL_JOYBUTTONUP:
			switch(event.jbutton.button) {
				case GP2X_BUTTON_X:
				gas=0;
				break;
				case GP2X_BUTTON_UP:
				v_dir=0;
				break;
				case GP2X_BUTTON_DOWN:
				v_dir=0;
				break;
				case GP2X_BUTTON_A:
				blow=0;
				break;
			}
			break;
			// other event types
		}
	}

	/* blow in smoke at high speed from the center-left.. */
	if (gas) {
		j=0;
		for (j=-4; j<=4; j+=4) {
			dens[IX(BORDER, BORDER+(HEIGHT/2)+j)]=F2X(3);
			/* beware, this depends very much on dt.. otherwise it blows right through the wall! */
		}
	}
	if (blow) {
		for (j=BORDER; j<HEIGHT+BORDER; j++) {
			u[IX(BORDER,j)]+=F2X(0.3);
			if (v_dir)
				v[IX(1,(HEIGHT/2)+j)]=F2X(v_dir)/8;
		}
	}

	return 0;
}

void Slock(SDL_Surface *screen)
{
	if ( SDL_MUSTLOCK(screen) ) {
		if ( SDL_LockSurface(screen) < 0 ) {
			return;
		}
	}
}

void Sulock(SDL_Surface *screen)
{
	if ( SDL_MUSTLOCK(screen) ) {
		SDL_UnlockSurface(screen);
	}
}

SDL_Surface *font;

void draw_text(int x, int y, const char* text, SDL_Surface *screen)
{
    SDL_Rect src;
    SDL_Rect dest;
    while (*text) {
        src.x = ((*text) & 0xF) * 6;
        src.y = (((*text) >> 4) - 2) * 8;
        src.w = 6;
        src.h = 8;
        dest.x = x;
        dest.y = y;
        SDL_BlitSurface(font, &src, screen, &dest);
        text++;
        x += 6;
    }
}

int main()
{
	int simulating=1, i=0, ret=0;
	int old_fr_count = 0;
	/* todo: examine these scalars */
	// 0.01, 0.01, 0.05
	freal diff=F2X(0.001), visc=F2X(0.0001), dt=F2X(.03);
	/* grid image buffer */
	unsigned char b[WIDTH*HEIGHT];
	/* output file name */
	char fname[256];

	long old_ticks=0, last_msg_ticks=0;

	#ifdef GP2X
	/* try to apply MMU hack. */
	int mmufd = open("/dev/mmuhack", O_RDWR);
	if (mmufd < 0) {
		system("/sbin/insmod mmuhack.o");
		mmufd = open("/dev/mmuhack", O_RDWR);
	}
	if (mmufd < 0) {
		fprintf(stderr, "MMU hack failed!");
		//goto end;
	} else {
		close(mmufd);
	}
	#endif

	printf("diff=%f visc=%f dt=%f\n", X2F(diff), X2F(visc), X2F(dt));

	char pgm_header[32];
	sprintf(pgm_header, "P5 %d %d 255\n", WIDTH, HEIGHT);

	for (i=0; i<SIZE; ++i) {
		dens_prev[i] = 0;
		u_prev[i] = 0;
		v_prev[i] = 0;
		dens[i] = 0;
		u[i] = 0;
		v[i] = 0;
	}
	log_dens_ptr=dens_prev;
	phys_dens_ptr=dens;

	printf("try to set bnd rout\n");
	model_set_bnd_rout(set_bnd_ball);

        /* initialise threads */
        if (init_threads()) {
                fprintf(stderr, "error initialising threads!\n");
        }
        printf("threading initialised\n");

	#if (!OUTPUT_IMAGES)

	/* SDL initialisation */
	atexit(SDL_Quit);

	if( SDL_Init(SDL_INIT_JOYSTICK|SDL_INIT_VIDEO /*|SDL_INIT_AUDIO */) <0 ) {
		printf("Unable to init SDL: %s\n", SDL_GetError());
		ret = -1;
		goto end;
	}
	SDL_JoystickOpen(0);
	SDL_ShowCursor(0);

	/* GP2X uses SWSURFACE? */
	screen = SDL_SetVideoMode(SDL_WIDTH, SDL_HEIGHT, 8, SDL_SWSURFACE /*|SDL_DOUBLEBUF */);
	if ( screen == NULL ) {
		printf("Unable to set %dX%d video: %s\n", SDL_WIDTH, SDL_HEIGHT, SDL_GetError());
		ret = -1;
		goto end;
	}

	#if 0
	Uint32 init = SDL_WasInit(SDL_INIT_AUDIO);
	if(init & SDL_INIT_AUDIO)
	{
		sound = 1;  // Audio init sucessful, use sound
	} else {
		sound = 0;  // Audio init unsucessful, don't use sound
	}
	#endif

	/* Fill colors with color information */
	for(i=0;i<256;i++) {
		colors[i].r=i;
		colors[i].g=i;
		colors[i].b=i;
	}

	/* Set palette */
	SDL_SetColors(screen, colors, 0, 256);

	SDL_Surface *temp = SDL_LoadBMP("TinyFont.bmp");
	if (temp == NULL) {
		printf("Unable to load bitmap: %s\n", SDL_GetError());
		return -1;
	}
	font = SDL_DisplayFormat(temp);
	/* clear screen*/
	SDL_FillRect(screen, NULL, 0);
	/* draw controls */
	draw_text(0, 230, " A: blow, X: smoke, up/down: direction, Start: exit", screen);

	#endif

	for (i=0; simulating; i++) {
		//printf("frame %d\n", i);

		SWAP(log_dens_ptr,phys_dens_ptr);

		if (get_from_ui_windtunnel(log_dens_ptr, u, v)) {
			simulating=0;
		}
		//printf("got from ui\n");

		vel_step(u,v,visc,dt);
		//printf("vel_step() done\n");

		dens_step(phys_dens_ptr,log_dens_ptr,u,v,diff,dt);
		dampen_ball(phys_dens_ptr);
		//printf("dens_step() done\n");

		#if OUTPUT_IMAGES

		//sprintf(fname, "grid%04d.gif", i);
		if (i>=199)
			simulating=0;

		sprintf(fname, "grid%04d.pgm", i);
		fout = fopen(fname, "wb");
		if (fout) {
			int x,y;

			//printf("output file '%s' opened\n", fname);
			fprintf(fout, pgm_header);
			//printf("wrote pgm header\n");

			for (x=BORDER; x<WIDTH+BORDER; x++) {
				for (y=BORDER; y<HEIGHT+BORDER; y++) {
					int h = X2I( 255*phys_dens_ptr[IX(x,y)] );
					b[(y-BORDER)*WIDTH+(x-BORDER)] = (h>255) ? 255 : h;
					//printf("%02X", b[x*N+y]);
					//fprintf(fout, "%4d", (int) b[x*N+y]);
				}
				//printf("x=%d\n",x);
				//fprintf(fout, "\n");
			}
			fwrite(b, 1, WIDTH*HEIGHT, fout);
			//printf("converted to bytes\n");

			fclose(fout);
		}

		#else

		Slock(screen);
	
		/* convert dens to frame buffer */
		Uint8 *pixels;
		int pitch;
		pixels = screen->pixels;
		pitch = screen->pitch;

		long ticks = SDL_GetTicks();
		if (old_ticks && (ticks-last_msg_ticks>5000)) {
			char fps_msg[10];
			//int t = ticks - old_ticks;
			//sprintf(fps_msg, "%d fps  ", 1000/t);
			int t = ticks - last_msg_ticks;
			int f = i - old_fr_count;
			sprintf(fps_msg, "%d fps  ", (f*1000)/t);
			draw_text(0, 0, fps_msg, screen);
			last_msg_ticks=ticks;
			old_fr_count=i;
		}
		old_ticks = ticks;

		int x,y;
		#if 0
		int dx=(SDL_WIDTH-WIDTH)/2, dy=(SDL_HEIGHT-HEIGHT)/2;
		for (y=BORDER; y<HEIGHT+BORDER; y++) {
			for (x=BORDER; x<WIDTH+BORDER; x++) {
				int h = X2I( 255*phys_dens_ptr[IX(x,y)] );
				if (h<0)
					h=0;
				else if (h>255)
					h=255;
				pixels[(y+dy-BORDER)*pitch + x+dx-BORDER] = h;
				//printf("%02X", b[x*N+y]);
			}
			//printf("x=%d\n",x);
		}
		#else
		int dx=(SDL_WIDTH-WIDTH*2)/2, dy=(SDL_HEIGHT-HEIGHT*2)/2;
		for (y=BORDER; y<HEIGHT+BORDER; y++) {
			for (x=BORDER; x<WIDTH+BORDER; x++) {
				int h = X2I( 255*phys_dens_ptr[IX(x,y)] );
				if (h<0)
					h=0;
				else if (h>255)
					h=255;
				pixels[(y*2+dy-BORDER)*pitch + x*2+dx-BORDER] = h;
				pixels[(y*2+dy-BORDER)*pitch + x*2+dx-BORDER +1] = h;
				pixels[(y*2+dy-BORDER+1)*pitch + x*2+dx-BORDER] = h;
				pixels[(y*2+dy-BORDER+1)*pitch + x*2+dx-BORDER +1] = h;
				//printf("%02X", b[x*N+y]);
			}
			//printf("x=%d\n",x);
		}
		#endif

		Sulock(screen);
		SDL_Flip(screen);

		#endif

	}

end:    cleanup_threads();

	SDL_Quit();  // Shut down the SDL library, freeing resources.
	
	#ifdef GP2X
	// This is how to quit back to the menu.  Just doing a return(0) will look like the GP2X hung.
	chdir("/usr/gp2x");
	execl("/usr/gp2x/gp2xmenu", "/usr/gp2x/gp2xmenu", NULL);
	// execl will go run the other program INSTEAD of this one.  
	#endif

	return 0;
}
