#include <stdio.h>
#include <string.h>

#include "model.h"

#define OUTPUT_IMAGES 1
#define SINGLE_PGM 0

/* velocity maps (u,v) t,t-1 */
static float u[SIZE], v[SIZE], u_prev[SIZE], v_prev[SIZE];
/* density maps t,t-1 */
static float dens[SIZE], dens_prev[SIZE];

FILE *fout;

/*---------------------------------------------------------------------------*/
/* boundary condition code                                                   */
/*---------------------------------------------------------------------------*/

#define BALL 1
#define D 9
#define D2 (D/2)
unsigned char ball[D*D]=
{
	0,0,0,0,1,0,0,0,0,
	0,0,1,1,1,1,1,0,0,
	0,1,1,1,1,1,1,1,0,
	0,1,1,1,1,1,1,1,0,
	1,1,1,1,1,1,1,1,1,
	0,1,1,1,1,1,1,1,0,
	0,1,1,1,1,1,1,1,0,
	0,0,1,1,1,1,1,0,0,
	0,0,0,0,1,0,0,0,0
};
unsigned char ball_occupied[D*D];

#define BIX(x,y) ((x)+D2+ D*((y)+D2))

void set_bnd_ball(int b, float *a)
{
	int i,x,y;

	for (i=1; i<=WIDTH; i++) {
		a[IX(i, 0)       ] = b==2 ? -a[IX(i,1)] : a[IX(i,1)];
		//a[IX(i, HEIGHT+1)] = b==2 ? -a[IX(i,HEIGHT)] : a[IX(i,HEIGHT)];
	}
	for (i=1; i<=HEIGHT; i++) {
		/* disable if we want a wind tunnel! */
		////a[IX(0   ,i)    ] = b==1 ? -a[IX(1,i)] : a[IX(1,i)];
		//a[IX(WIDTH+1 ,i)] = b==1 ? -a[IX(WIDTH,i)] : a[IX(WIDTH,i)];
	}

	/* basin corners */
	//a[IX(0  ,0  )         ] = .5*(a[IX(1,0  )         ]+a[IX(0  ,1)]);
	//a[IX(0  ,HEIGHT+1)    ] = .5*(a[IX(1,HEIGHT+1)    ]+a[IX(0  ,HEIGHT)]);
	//a[IX(WIDTH+1,0  )     ] = .5*(a[IX(WIDTH,0  )     ]+a[IX(WIDTH+1,1)]);
	//a[IX(WIDTH+1,HEIGHT+1)] = .5*(a[IX(WIDTH,HEIGHT+1)]+a[IX(WIDTH+1,HEIGHT)]);

	#if BALL
	/* an object blocking the stream.. */
	memset(ball_occupied, 0, D*D);
	int xoff=(WIDTH/8), yoff=(HEIGHT/2);

	/* build left and right walls */
	for (y=-D2; y<=D2; y++) {
		int in=0;
		for(x=-D2; x<=D2; x++) {
			int bool=ball[BIX(x,y)];
			if (bool && !in) {
				/* make a left wall */
				a[IX(x+xoff,y+yoff)]= b&1 ? -a[IX(x+xoff-1,y+yoff)] : a[IX(x+xoff-1,y+yoff)];
				in=1;
				ball_occupied[BIX(x,y)]=1;
			} else if (!bool && in) {
				/* make a right wall */
				a[IX(x+xoff,y+yoff)]= b&1 ? -a[IX(x+xoff+1,y+yoff)] : a[IX(x+xoff+1,y+yoff)];
				in=0;
				ball_occupied[BIX(x,y)]=1;
			}
		}
	}
	/* build top and bottom walls */
	for (x=-D2; x<=D2; x++) {
		int in=0;
		for(y=-D2; y<=D2; y++) {
			int bool=ball[BIX(x,y)];

			if (!ball_occupied[BIX(x,y)]) {
				if (bool && !in) {
					/* make a top wall */
					a[IX(x+xoff,y+yoff)]= b&2 ? -a[IX(x+xoff,y+yoff-1)] : a[IX(x+xoff,y+yoff-1)];
					in=1;
				} else if (!bool && in) {
					/* make a bottom wall */
					a[IX(x+xoff,y+yoff)]= b&2 ? -a[IX(x+xoff,y+yoff+1)] : a[IX(x+xoff,y+yoff+1)];
					in=0;
				}
			}

		}

	}
	#endif /*#if BALL*/

}

void get_from_ui(float *dens_prev, float *u_prev, float *v_prev, int i)
{
	float speedhi=40, speedlo=16;

	if (i<40) {
		dens_prev[IX(WIDTH/2,    HEIGHT/2)    ]+=2;
		dens_prev[IX(WIDTH/2,    (HEIGHT/2)+1)]+=2;
		dens_prev[IX((WIDTH/2)+1,HEIGHT/2)    ]+=2;
		dens_prev[IX((WIDTH/2)+1,(HEIGHT/2)+1)]+=2;
	}
	if (i<40) {
		u_prev[IX(WIDTH/2,    HEIGHT/2)    ]-=speedhi;
		u_prev[IX(WIDTH/2,    (HEIGHT/2)+1)]-=speedhi;
		u_prev[IX((WIDTH/2)+1,HEIGHT/2)    ]-=speedhi;
		u_prev[IX((WIDTH/2)+1,(HEIGHT/2)+1)]-=speedhi;
		v_prev[IX(WIDTH/2,    HEIGHT/2)    ]+=speedlo;
		v_prev[IX(WIDTH/2,    (HEIGHT/2)+1)]+=speedlo;
		v_prev[IX((WIDTH/2)+1,HEIGHT/2)    ]+=speedlo;
		v_prev[IX((WIDTH/2)+1,(HEIGHT/2)+1)]+=speedlo;
	}
	if ((i>=120) && (i<150)) {
		v_prev[IX(WIDTH/2,    HEIGHT/2)    ]+=speedhi;
		v_prev[IX(WIDTH/2,    (HEIGHT/2)+1)]+=speedhi;
		v_prev[IX((WIDTH/2)+1,HEIGHT/2)    ]+=speedhi;
		v_prev[IX((WIDTH/2)+1,(HEIGHT/2)+1)]+=speedhi;
	}

}

void get_from_ui_windtunnel(float *dens_prev, float *u_prev, float *v_prev, int fr)
{
	int i,j,x,y;

	/* blow in smoke at high speed from the center-left.. */
		j=0;
	for (j=-4; j<=4; j+=4) {
		dens_prev[IX(1,(HEIGHT/2)+j)]+=2;
		/* beware, this depends very much on dt.. otherwise it blows right through the wall! */
	}
	for (j=-(HEIGHT/2); j<=(HEIGHT/2); j++) {
		u_prev[IX(1,(HEIGHT/2)+j)]+=5;
		//v_prev[IX(1,(HEIGHT/2)+j)]=0;
	}

}

void init_clouds(float *dens_prev)
{
	int i,j;

#if 0
	for (j=-8+8; j<+8+8; j++) {
		for (i=2; i<WIDTH-1; i++) {
			dens_prev[IX(i,(HEIGHT/2)+j)]=2;
		}
	}
#endif
#if 0
	/* blow in smoke at high speed from the center-left.. */
	for (j=-(HEIGHT/2); j<=0; j++) {
		for (i=2; i<WIDTH-1; i++) {
			u_prev[IX(i,(HEIGHT/2)+j)]=5;
		}
		//v_prev[IX(1,(HEIGHT/2)+j)]=0;
	}
	for (j=1; j<=(HEIGHT/2); j++) {
		for (i=2; i<WIDTH-1; i++) {
			u_prev[IX(i,(HEIGHT/2)+j)]=.5;
		}
		//v_prev[IX(1,(HEIGHT/2)+j)]=0;
	}
#endif
}

void get_from_ui_layers(float *dens_prev, float *u_prev, float *v_prev, int fr)
{
	int i,j,x,y;
#if 0
	/* blow in smoke at high speed from the center-left.. */
	for (j=-(HEIGHT/2); j<=0; j++) {
		for (i=2; i<WIDTH-1; i+=WIDTH/9) {
			u_prev[IX(i,(HEIGHT/2)+j)]+=1;
		}
		//v_prev[IX(1,(HEIGHT/2)+j)]=0;
	}
	for (j=1; j<=(HEIGHT/2); j++) {
		for (i=2; i<WIDTH-1; i+=WIDTH/8) {
			u_prev[IX(i,(HEIGHT/2)+j)]-=1;
		}
		//v_prev[IX(1,(HEIGHT/2)+j)]=0;
	}
#else
	for (j=-16; j<=0; j++) {
		dens_prev[IX(1,(HEIGHT/2)+j)]+=2;
		/* beware, this depends very much on dt.. otherwise it blows right through the wall! */
	}
	/* blow in smoke at high speed from the center-left.. */
	for (j=-(HEIGHT/2); j<=0; j++) {
		u_prev[IX(1,(HEIGHT/2)+j)]+=20;
		//v_prev[IX(1,(HEIGHT/2)+j)]=0;
	}
	for (j=1; j<=(HEIGHT/2); j++) {
		u_prev[IX(1,(HEIGHT/2)+j)]+=1;
		//v_prev[IX(1,(HEIGHT/2)+j)]=0;
	}
#endif
}

void write_bytes(void *user, int len, char *bytes)
{
	fwrite(bytes, len, 1, fout);
}

int main()
{
	int simulating=1, i=0;
	/* todo: examine these scalars */
	float diff=.001, visc=.0001, dt=.05;
	/* grid palette */
	unsigned char palette[256*3];
	/* grid image buffer */
	unsigned char b[WIDTH*HEIGHT];
	/* output file name */
	char fname[256];

	char pgm_header[32];
	sprintf(pgm_header, "P5 %d %d 255\n", WIDTH, HEIGHT);

	/* create palette */
	for (i=0; i<256; i++) {
		palette[i*3]=i;
		palette[i*3+1]=i;
		palette[i*3+2]=i;
	}

	memset(dens_prev, 0, SIZE*sizeof(float));
	memset(u_prev, 0, SIZE*sizeof(float));
	memset(v_prev, 0, SIZE*sizeof(float));

	printf("try to set bnd rout\n");

	model_set_bnd_rout(set_bnd_ball);

	#if (SINGLE_PGM && OUTPUT_IMAGES)
	sprintf(fname, "gridanim.pgm");
	fout = fopen(fname, "wb");
	#endif

	for (i=0; simulating; i++) {
		printf("iteration %d\n", i);
		//printf(".");
		if (i>=999)
			simulating=0;

		get_from_ui_windtunnel(dens_prev, u_prev, v_prev, i);
		//get_from_ui_layers(dens_prev, u_prev, v_prev, i);
		//printf("got from ui\n");

		vel_step(u,v,u_prev,v_prev,visc,dt);
		//printf("vel_step() done\n");
		dens_step(dens,dens_prev,u,v,diff,dt);
		//printf("dens_step() done\n");

		//printf("d=%f (u,v)=(%f,%f)\n", dens[IX(N/2,N/3)], u[IX(N/2,N/3)], v[IX(N/2,N/3)]);
		//draw_dens(n,dens);

		//sprintf(fname, "grid%04d.gif", i);

		#if OUTPUT_IMAGES

		#if (!SINGLE_PGM)
		sprintf(fname, "grid%04d.pgm", i/2);
		fout = fopen(fname, "wb");
		#endif

		if (fout) {
			int x,y;

			printf("output file '%s' opened\n", fname);
			fprintf(fout, pgm_header);
			//printf("wrote pgm header\n");

			for (x=1; x<=WIDTH; x++) {
				for (y=1; y<=HEIGHT; y++) {
					int h=255.0*dens[IX(x,y)];
					b[(y-1)*WIDTH+(x-1)] = (h>255) ? 255 : h;
					//printf("%02X", b[x*N+y]);
					//fprintf(fout, "%4d", (int) b[x*N+y]);
				}
				//printf("x=%d\n",x);
				//fprintf(fout, "\n");
			}
			fwrite(b, 1, WIDTH*HEIGHT, fout);

			//printf("converted to bytes\n");

			#if (!SINGLE_PGM)
			fclose(fout);
			#else
			fprintf(fout, "\n");
			#endif

		}
		#endif

	}

	#if (OUTPUT_IMAGES && SINGLE_PGM)
	fclose(fout);
	#endif

	return 0;
}
