#include <string.h>
#include <math.h>

#include "model.h"

void (*set_bnd)(int b, float *a);
void model_set_bnd_rout(void *func)
{
	printf("bnd_rout (before) =0x%08X\n", set_bnd);
	printf("void ptr argument=0x%08X\n", func);
	set_bnd=func;
	printf("bnd_rout (after) =0x%08X\n", set_bnd);
}
//void set_bnd(int b, float *a) {}

void print_min_max(char *name, float *field)
{
	#if DEBUG
	int i;
	float min=+10e40, max=0;
	for (i=0; i<SIZE; ++i) {
		float a = fabs(field[i]);
		if (a > max) {
			max = a;
		}
		if ((a < min) && (a != 0.0)) {
			min = a;
		}
	}
	printf("%s: min=%e, max=%e\n", name, min, max);
	#endif
}
	
/* apply external density influence */
void add_source(float *x, float *s, float dt)
{
	int i;
	for (i=0; i<SIZE; i++) x[i]+=dt*s[i];
}

void diffuse_bad(int b, float *x, float *x0, float diff, float dt)
{
	int i,j;
	float a=dt*diff*WIDTH*HEIGHT;

	for (i=1; i<=WIDTH; i++) {
		for (j=1; j<=HEIGHT; j++) {
			x[IX(i,j)] = x0[IX(i,j)]
			           + a*( x0[IX(i-1,j)]+x0[IX(i+1,j)]+x0[IX(i,j-1)]+x0[IX(i,j+1)]-4*x0[IX(i,j)] );
		}
	}
	set_bnd(b, x);
}

/* gauss seidel stabilised.. still grows too much in some cases :/ */
void diffuse(int b, float *x, float *x0, float diff, float dt)
{
	int i, j, k;
	float a=dt*diff*WIDTH*HEIGHT;

	for (k=0; k<20; k++) {
		for (i=1; i<=WIDTH; i++) {
			for (j=1; j<=HEIGHT; j++) {
				x[IX(i,j)] = (
			                 x0[IX(i,j)]
				             + a*(x[IX(i-1,j)]+x[IX(i+1,j)]+x[IX(i,j-1)]+x[IX(i,j+1)])
							 )
						     / (1+4*a);
			}
		}
		set_bnd(b, x);
	}
}

void diffuse_fixed(int b, float *x, float *x0, float diff, float dt)
{
	int i, j, k;
	float a=dt*diff*WIDTH*HEIGHT;

	for (i=1; i<=WIDTH; i++) {
		for (j=1; j<=HEIGHT; j++) {
			x[IX(i,j)] = (
		                 x0[IX(i,j)]
			             + a*(x0[IX(i-1,j)]+x0[IX(i+1,j)]+x0[IX(i,j-1)]+x0[IX(i,j+1)])
						 )
					     / (1+4*a);
		}
	}
	set_bnd(b, x);
}

void advect(int b, float *d, float *d0, float *u, float *v, float dt)
{
	int i,j,i0,j0,i1,j1;
	float x,y,s0,t0,s1,t1,dt0;

	dt0=dt* ((float) (WIDTH+HEIGHT))*0.5;
	for (i=1; i<=WIDTH; i++) {
		for (j=1; j<=HEIGHT; j++) {
			/* backtrack to previous grid position.. */
			x = i-dt0*u[IX(i,j)];
			y = j-dt0*v[IX(i,j)];

			if (x<.5)
				x=.5;
		    else if (x>WIDTH+.5)
			    x=WIDTH+.5;
		    i0=(int)x;
		    i1=i0+1;

			if (y<.5)
			    y=.5;
		    else if (y>HEIGHT+.5)
			    y=HEIGHT+.5;
		    j0=(int)y;
		    j1=j0+1;

			s1=x-i0;
			s0=1-s1;

			t1=y-j0;
			t0=1-t1;

			/* bilinear interpolation.. */
			d[IX(i,j)] = s0*(t0*d0[IX(i0,j0)]+t1*d0[IX(i0,j1)])
			           + s1*(t0*d0[IX(i1,j0)]+t1*d0[IX(i1,j1)]);
		}
	}
}

void dens_step(float *x, float *x0, float *u, float *v, float diff, float dt)
{
	////add_source(x,x0,dt);
	//diffuse_fixed(0,x,x0,diff,dt);
	/* no diffusion, the advection has enough of that.. (bilinear interpolation (!)) */
	int i, size=SIZE;
	for (i=0; i<size; i++) x[i]=x0[i];

	advect(0,x0,x,u,v,dt);
	print_min_max("'dens' after advect()", x0);

}

void project(float *u, float *v, float *p, float *div)
{
	int i,j,k;
	float h;

	h=2.0/((float)(HEIGHT+WIDTH));
	for (i=1; i<=WIDTH; i++) {
		for (j=1; j<=HEIGHT; j++) {
			div[IX(i,j)] = -.5*h*(u[IX(i+1,j)]-u[IX(i-1,j)]+v[IX(i,j+1)]-v[IX(i,j-1)]);
			p[IX(i,j)]=0;
		}
	}
	set_bnd(0,div);
	set_bnd(0,p);

	for (k=0; k<20; k++) {
		for (i=1; i<=WIDTH; i++) {
			for (j=1; j<=HEIGHT; j++) {
				p[IX(i,j)] = (
				             div[IX(i,j)]
						   + p[IX(i-1,j)]+p[IX(i+1,j)]+p[IX(i,j-1)]+p[IX(i,j+1)]
				             )/4;
			}
		}
		set_bnd(0,p);
	}

	for (i=1; i<=WIDTH; i++) {
		for (j=1; j<=HEIGHT; j++) {
				u[IX(i,j)] -= .5*(p[IX(i+1,j)]-p[IX(i-1,j)])/h;
				v[IX(i,j)] -= .5*(p[IX(i,j+1)]-p[IX(i,j-1)])/h;
		}
	}
	set_bnd(1,u);
	set_bnd(2,v);
}

void vel_step(float *u, float *v, float *u0, float *v0, float visc, float dt)
{
	add_source(u,u0,dt);
	add_source(v,v0,dt);

	print_min_max("'u' after add_source()", u);
	print_min_max("'v' after add_source()", v);

	SWAP(u0,u);
	SWAP(v0,v);

	diffuse_fixed(1,u,u0,visc,dt);
	diffuse_fixed(2,v,v0,visc,dt);

	print_min_max("'u' after diffuse_fixed()", u);
	print_min_max("'v' after diffuse_fixed()", v);

	project(u,v,u0,v0);
	print_min_max("'u' after project(1)", u);
	print_min_max("'v' after project(1)", v);

	SWAP(u0,u);
	SWAP(v0,v);

	advect(1,u,u0,u0,v0,dt);
	advect(2,v,v0,u0,v0,dt);
	print_min_max("'u' after advect()", u);
	print_min_max("'v' after advect()", v);

	project(u,v,u0,v0);
	print_min_max("'u' after project(2)", u);
	print_min_max("'v' after project(2)", v);
}

