/*
 * 'safe' multi-thread version using:
 * - producer-consumer scheme (only mutexes, no signals)
 * - only two threads (the main one does everything and bottom part of 
 *   projection, the additional thread does top part of projection)
 */

#include <stdio.h>
#include <pthread.h>
#include <string.h>

#include "model_int_threads.h"

#define DEBUG 0

void (*set_bnd)(int b, freal *a);
void model_set_bnd_rout(void *func)
{
	printf("bnd_rout (before) =0x%08X\n", set_bnd);
	printf("void ptr argument=0x%08X\n", func);
	set_bnd=func;
	printf("bnd_rout (after) =0x%08X\n", set_bnd);
}
//void set_bnd(int b, freal *a) {}

/* apply external density influence */
void add_source(freal *x, freal *s, freal dt)
{
	int i,j;

	#if 0
	//dt=X2I(dt<<10);
	//for (i=SIZE-1; i>=0; --i) x[i]+=(dt*s[i])>>10;
	for (i=SIZE-1; i>=0; --i) x[i]+=s[i]>>6;
	#else
	for (i=SIZE-1; i>=0; --i) x[i]+= XM(dt,s[i]);
	#endif

	#if DEBUG
	printf("add_source()\n");
	for (j=0; j<=HEIGHT+1; j++) {
		printf("s = ");
		for (i=0; i<=WIDTH+1; i++) {
			printf("%04X ", s[IX(i,j)]);
		}
		printf("\n");
	}
	printf("\n");

	for (j=0; j<=HEIGHT+1; j++) {
		printf("x = ");
		for (i=0; i<=WIDTH+1; i++) {
			printf("%04X ", x[IX(i,j)]);
		}
		printf("\n");
	}
	printf("\n");
	#endif
}

void diffuse_bad(int b, freal *x, freal *x0, freal diff, freal dt)
{
	int i,j;
	freal a=XM(dt,diff)*WIDTH*HEIGHT;

	for (i=BORDER; i<WIDTH+BORDER; i++) {
		for (j=BORDER; j<HEIGHT+BORDER; j++) {
			x[IX(i,j)] = x0[IX(i,j)]
			           + XM(a, ( x0[IX(i-1,j)]+x0[IX(i+1,j)]+x0[IX(i,j-1)]+x0[IX(i,j+1)]-4*x0[IX(i,j)] ));
		}
	}
	set_bnd(b, x);
}

/* gauss seidel stabilised.. still grows too much in some cases :/ */
void diffuse(int b, freal *x, freal *x0, freal diff, freal dt)
{
	int i, j, k;
	freal a=XM(dt*WIDTH,diff)*HEIGHT;

	for (k=0; k<20; k++) {
		for (j=BORDER; j<HEIGHT+BORDER; j++) {
			for (i=BORDER; i<WIDTH+BORDER; i++) {
				x[IX(i,j)] = (
			                       X1_0 * x0[IX(i,j)]
				             + a * (x[IX(i-1,j)]+x[IX(i+1,j)]+x[IX(i,j-1)]+x[IX(i,j+1)])
					     )
					     / (X1_0+4*a);
			}
		}
		set_bnd(b, x);
	}
}

void diffuse_fixed(int b, freal *x, freal *x0, freal diff, freal dt)
{
	int i, j, k;
	freal a=XM(dt*WIDTH,diff)*HEIGHT;
	long s1 = ((long long) X1_0<<12)/((X1_0+4*a));
	long s2 = ((long long) a<<12)/((X1_0+4*a));

	for (j=BORDER; j<HEIGHT+BORDER; j++) {
		for (i=BORDER; i<WIDTH+BORDER; ++i) {
			#if FIXEDPOINT
			x[IX(i,j)] =
			(
			  s1*(x0[IX(i,j)]>>4) + s2*((x0[IX(i-1,j)]+x0[IX(i+1,j)]+x0[IX(i,j-1)]+x0[IX(i,j+1)])>>4)
			)>>8;
			#else
			x[IX(i,j)] = (
				       XML( X1_0, x0[IX(i,j)] )
				     + XML( a, (x0[IX(i-1,j)]+x0[IX(i+1,j)]+x0[IX(i,j-1)]+x0[IX(i,j+1)]) )
				     )
				     / (X1_0+4*a);
			#endif
		}
	}
	set_bnd(b, x);

	#if DEBUG
	printf("diffuse_fixed(dt=0x%08X * diff=0x%08X = a=0x%08X)\n", dt, diff, a);
	for (j=0; j<=HEIGHT+1; j++) {
		printf("x0 = ");
		for (i=0; i<=WIDTH+1; i++) {
			printf("%04X ", x0[IX(i,j)]);
		}
		printf("\n");
	}
	printf("\n");

	for (j=0; j<=HEIGHT+1; j++) {
		printf("x = ");
		for (i=0; i<=WIDTH+1; i++) {
			printf("%04X ", x[IX(i,j)]);
		}
		printf("\n");
	}
	printf("\n");
	#endif
}

void advect(int b, freal *d, freal *d0, freal *u, freal *v, freal dt)
{
	int i,j,i0,j0,i1,j1;
	freal x,y,s0,t0,s1,t1,dt0;

	//printf("d=0x%08X d0=0x%08X u=0x%08X v=0x%08X\n", d, d0, u, v);

	dt0=HALVE( (dt * (WIDTH+HEIGHT)) );
	//printf("dt0=0x%X\n", dt0);
	for (j=BORDER; j<HEIGHT+BORDER; j++) {
		//printf("i=%d\n", i);

		for (i=BORDER; i<WIDTH+BORDER; i++) {
			/* backtrack to previous grid position.. */
			//printf("j=%d\t", j);
			x = I2X(i) - XM(dt0, u[IX(i,j)]);
			y = I2X(j) - XM(dt0, v[IX(i,j)]);

			unsigned long xs = (unsigned long) (x-HALVE(X1_0));
			if (xs > I2X(WIDTH)) {
				if ((long) xs < 0)
					x = HALVE(X1_0);
				else if (xs > I2X(WIDTH))
					x=I2X(WIDTH)+HALVE(X1_0);
			}
			i0=X2I(x);
			i1=i0+1;

			#if 1
			unsigned long ys = (unsigned long) (y-HALVE(X1_0));
			if (ys > I2X(WIDTH)) {
				if ((long) ys < 0)
					y = HALVE(X1_0);
				else if (ys > I2X(HEIGHT))
					y=I2X(HEIGHT)+HALVE(X1_0);
			}
			#else
			if (y < HALVE(X1_0))
				y = HALVE(X1_0);
			else if ( y > I2X(HEIGHT)+HALVE(X1_0) )
				y = I2X(HEIGHT)+HALVE(X1_0);
			#endif
			j0=X2I(y);
			j1=j0+1;

			s1=x-I2X(i0);
			s0=X1_0-s1;
			//printf("s0=0x%X ", s0);

			t1=y-I2X(j0);
			t0=X1_0-t1;
			//printf("t0=0x%X ", t0);

			/* bilinear interpolation.. */
			d[IX(i,j)] = XM(s0, (XM(t0, d0[IX(i0,j0)]) + XM(t1, d0[IX(i0,j1)])) )
			           + XM(s1, (XM(t0, d0[IX(i1,j0)]) + XM(t1, d0[IX(i1,j1)])) );
		}
	}
	//set_bnd(b, d);
}

/* bilinear interpolation.. */
#define INTERP() \
i0=X2I(x); \
i1=i0+1; \
j0=X2I(y); \
j1=j0+1; \
s1=(x-I2X(i0))>>11; \
s0=0x100-s1; \
t1=(y-I2X(j0))>>11; \
t0=0x100-t1; \
d[IX(i,j)] = \
( \
  s0 * ((t0 * d0[IX(i0,j0)] + t1 * d0[IX(i0,j1)])>>8 ) \
+ s1 * ((t0 * d0[IX(i1,j0)] + t1 * d0[IX(i1,j1)])>>8 ) \
)>>8;
//i0=X2I(x); \
//i1=i0+1; \
//j0=X2I(y); \
//j1=j0+1; \
//s1=x-I2X(i0); \
//s0=X1_0-s1; \
//t1=y-I2X(j0); \
//t0=X1_0-t1; \
//d[IX(i,j)] = \
//  XM(s0, (XM(t0, d0[IX(i0,j0)]) + XM(t1, d0[IX(i0,j1)])) ) \
//+ XM(s1, (XM(t0, d0[IX(i1,j0)]) + XM(t1, d0[IX(i1,j1)])) );

#define BACKTRACK() \
x = I2X(i) - dt0 * u[IX(i,j)]; \
y = I2X(j) - dt0 * v[IX(i,j)];

/*
 * optimised version with respect to clipping.
 * clipping is done only at the side pixels..
 * under the assumption that a cell cannot move more than one unit per frame..
 * this is reasonable because this is also a criterion for every boundary condition.
 * this is significantly faster..
 */
void advect_new(int b, freal *d, freal *d0, freal *u, freal *v, freal dt)
{
	int i,j,i0,j0,i1,j1;
	freal x,y,s0,t0,s1,t1,dt0;
	//printf("d=0x%08X d0=0x%08X u=0x%08X v=0x%08X\n", d, d0, u, v);
	/* dt0 is big enough and can be respresented as an int.. */
	dt0=X2I(HALVE( dt * (WIDTH+HEIGHT) ));

	/* top case.. check for out of bounds extrapolation.. */
	i=BORDER;
	j=BORDER;
	BACKTRACK();
	if (x < HALVE(X1_0))
		x = HALVE(X1_0);
	if (y < HALVE(X1_0))
		y = HALVE(X1_0);
	INTERP();
	for (i=BORDER+1; i<WIDTH+BORDER-1; i++) {
		/* backtrack to previous grid position.. */
		BACKTRACK();
		if (y < HALVE(X1_0))
			y = HALVE(X1_0);
		INTERP();
	}
	BACKTRACK();
	if (x > I2X(WIDTH)+HALVE(X1_0))
		x=I2X(WIDTH)+HALVE(X1_0);
	if (y < HALVE(X1_0))
		y = HALVE(X1_0);
	INTERP();

	/* body case.. don't check for out of bounds extrapolation in Y dir.. */
	for (j=BORDER+1; j<HEIGHT+BORDER-1; j++) {
		i=BORDER;
		BACKTRACK();
		if (x < HALVE(X1_0))
			x = HALVE(X1_0);
		INTERP();

		for (i=BORDER+1; i<WIDTH+BORDER-1; i++) {
			/* backtrack to previous grid position.. */
			BACKTRACK();
			INTERP();
		}

		BACKTRACK();
		if (x > I2X(WIDTH)+HALVE(X1_0))
			x=I2X(WIDTH)+HALVE(X1_0);
		INTERP();
	}

	/* bottom case.. check for out of bounds extrapolation.. */
	i=BORDER;
	BACKTRACK();
	if (x < HALVE(X1_0))
		x = HALVE(X1_0);
	if (y > I2X(HEIGHT)+HALVE(X1_0))
		y = I2X(HEIGHT)+HALVE(X1_0);
	INTERP();
	for (i=BORDER+1; i<WIDTH+BORDER-1; i++) {
		/* backtrack to previous grid position.. */
		BACKTRACK();
		if (y > I2X(HEIGHT)+HALVE(X1_0))
			y = I2X(HEIGHT)+HALVE(X1_0);
		INTERP();
	}
	BACKTRACK();
	if (x > I2X(WIDTH)+HALVE(X1_0))
		x=I2X(WIDTH)+HALVE(X1_0);
	if (y > I2X(HEIGHT)+HALVE(X1_0))
		y = I2X(HEIGHT)+HALVE(X1_0);
	INTERP();

	//set_bnd(b, d);
}

freal tmpbuf[SIZE];

/* i just don't get the add source in this case.. it only works when x0 (prev_dens) is cleared except for new sources */
void dens_step(freal *x, freal *x0, freal *u, freal *v, freal diff, freal dt)
{
	advect_new(0,x,x0,u,v,dt);
}

#define GS_NUM_LOOPS 20

void project_gauss_seidel(int start_line, int n_lines, freal *div, freal *p)
{
        int i,j,end;

        end = n_lines+start_line;
        for (j = BORDER+start_line; j < end+BORDER; j++) {
                int y_offset=j*(WIDTH+BORDER*2);
                for (i=BORDER; i<WIDTH+BORDER; i++) {
                        p[IX(i,j)] = (
                                        div[IX(i,j)]
                                        + p[IX(i-1,j)]+p[IX(i+1,j)]+p[IX(i,j-1)]+p[IX(i,j+1)]
                                        )/4;
                }
        }
        set_bnd(0,p);
}

freal p[SIZE], div[SIZE];

pthread_t top_worker_thread;
pthread_t btm_worker_thread;
pthread_mutex_t empty_mutex, full_mutex;
static int thread_cmd_state;
static int run_threads;
float *map_u, *map_v;
int n_work_iters;
int workers_done, boss_done;

float *top_map_ptr, *btm_map_ptr;

void *work_top(void *idp)
{
        int i, j, top_iters;

        printf("work_top() started\n");

        while(run_threads) {

		//printf("run_threads=%d\n", run_threads);
                pthread_mutex_unlock(&full_mutex);

                for (top_iters=0; (top_iters<n_work_iters) && run_threads; top_iters++) {
        
                        /* perform the bulk of the work.. */
                        project_gauss_seidel(0, (HEIGHT/2)-1, div, p);
        
                        /* keep the center range of the map locked */
			//printf("work_top(): try down(empty)\n");
                        pthread_mutex_lock(&empty_mutex);
			//printf("work_top(): empty mutex locked\n");
        
                        /* perform work on the center range */
                        project_gauss_seidel((HEIGHT/2)-1, 1, div, p);
        
                        /* unlock the center range */
			//printf("work_top(): try up(full)\n");
                        pthread_mutex_unlock(&full_mutex);
			//printf("work_top(): done up(full)\n");
        
                        //printf("top: iteration %d done.\n", top_iters);
                }

        }

        printf("top: run_threads==0\n");

        pthread_exit(NULL);
}

pthread_attr_t attr;

int init_threads()
{
        int ret;

	printf("init_threads(): (W,H) = (%d,%d)\n", WIDTH, HEIGHT);

        /* initialise map */
        //memset(map, 0, sizeof(float)*MAP_SIZE);

        /* Initialize mutex and condition variable objects */
        pthread_mutex_init(&empty_mutex, NULL);
        pthread_mutex_init(&full_mutex, NULL);

        /*
        For portability, explicitly create threads in a joinable state 
        */
        pthread_attr_init(&attr);
        ret = pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
        if (ret)
                return ret;

        /* set threads to default state (alive) */
        run_threads = 1;

        ret = pthread_create(&top_worker_thread, &attr, work_top, (void *) 0);
        if (ret)
                return ret;

        pthread_mutex_lock(&full_mutex);

        //pthread_create(&btm_worker_thread, &attr, work_bottom, (void *) 1);
        //if (ret)
        //        return ret;

        return 0;
}

int provide_work(int n_iters, float *map_u_ptr, float *map_v_ptr)
{
        n_work_iters = n_iters;
        map_u = map_u_ptr;
        map_v = map_v_ptr;
}

int cleanup_threads()
{
        printf("cleanup_threads()\n");

        /* lower flag for threads.. */
        run_threads = 0;

	printf("cleanup threads.. try \n");
	pthread_mutex_unlock(&empty_mutex);
	pthread_mutex_lock(&full_mutex);
	printf("cleanup threads.. done try \n");

        pthread_join(top_worker_thread, NULL);
        printf("main() top thread joined\n");

        /* Clean up and exit */
        pthread_attr_destroy(&attr);

        //pthread_exit (NULL);
}

#define PARALLEL_GS 1

void project(freal *u, freal *v, int gs_loops)
{
	int i,j,k,m;
	freal h;

	#if 1
	for (i=0; i<WIDTH+BORDER*2; ++i) {
		p[IX(i,0)]=0;
		p[IX(i,HEIGHT+BORDER)]=0;
		div[IX(i,0)]=0;
		div[IX(i,HEIGHT+BORDER)]=0;
	}
	for (j=0; j<HEIGHT+BORDER*2; ++j) {
		p[IX(0,j)]=0;
		p[IX(WIDTH+BORDER,j)]=0;
		div[IX(0,j)]=0;
		div[IX(WIDTH+BORDER,j)]=0;
	}
	#endif

	h = I2X(2)/(HEIGHT+WIDTH);
	long sc = -h/2;
	for (j=BORDER; j<HEIGHT+BORDER; j++) {
		for (i=BORDER; i<WIDTH+BORDER; i++) {
			div[IX(i,j)] = XM( sc, (u[IX(i+1,j)]-u[IX(i-1,j)]+v[IX(i,j+1)]-v[IX(i,j-1)]) );
			p[IX(i,j)]=0;
		}
	}
	set_bnd(0,div);
	set_bnd(0,p);

        /* perform Gauss-Seidel iterations */
        #if PARALLEL_GS

        n_work_iters = gs_loops;

        /* wait for workers to finish (blocked-waiting) */
	int btm_iters;
	for (btm_iters=0; btm_iters<n_work_iters; btm_iters++) {

		/* wait for top worker to complete its job (the center range should be free) */
		//printf("work_bottom(): try to lock\n");
		pthread_mutex_lock(&full_mutex);
		//printf("work_bottom(): full mutex locked\n");

		/* perform work using the center range */
		project_gauss_seidel(HEIGHT/2, 1, div, p);

		/* unlock the center range */
		pthread_mutex_unlock(&empty_mutex);

		/* perform the bulk of the work.. */
		project_gauss_seidel((HEIGHT/2)+1, (HEIGHT/2)-1, div, p);

		//printf("btm: iteration %d done.\n", btm_iters);
	}

	//boss_done = 1;
        //printf("boss: good work boys!\n");

        #else
	for (k=0; k<gs_loops; k++) {
                //project_gauss_seidel(0, HEIGHT, div, p);
                project_gauss_seidel(0, HEIGHT/2, div, p);
                project_gauss_seidel(HEIGHT/2, HEIGHT/2, div, p);
	}
        #endif

	freal inv = XDL(X1_0, 2*h);
	long s = (WIDTH+HEIGHT)>>2;
	for (j=BORDER; j<HEIGHT+BORDER; j++) {
		for (i=BORDER; i<WIDTH+BORDER; i++) {
			//u[IX(i,j)] -= XM(p[IX(i+1,j)]-p[IX(i-1,j)], inv);
			//v[IX(i,j)] -= XM(p[IX(i,j+1)]-p[IX(i,j-1)], inv);
			u[IX(i,j)] -= s*(p[IX(i+1,j)]-p[IX(i-1,j)]);
			v[IX(i,j)] -= s*(p[IX(i,j+1)]-p[IX(i,j-1)]);
		}
	}
	set_bnd(1,u);
	set_bnd(2,v);
}

freal tmpu[SIZE], tmpv[SIZE];

void vel_step(freal *u, freal *v, freal visc, freal dt)
{
	#if DEBUG
	printf("vel_step(visc=0x%08X)\n", visc);
	#endif

	diffuse_fixed(1,tmpu,u,visc,dt);
	diffuse_fixed(2,tmpv,v,visc,dt);
	project(tmpu,tmpv,50);
	advect_new(1,u,tmpu,tmpu,tmpv,dt);
	advect_new(2,v,tmpv,tmpu,tmpv,dt);
	project(u,v,50);
}
