#include <pthread.h>
#include <string.h>

#include "model_int_threads.h"

#define DEBUG 0

void (*set_bnd)(int b, freal *a);
void model_set_bnd_rout(void *func)
{
	printf("bnd_rout (before) =0x%08X\n", set_bnd);
	printf("void ptr argument=0x%08X\n", func);
	set_bnd=func;
	printf("bnd_rout (after) =0x%08X\n", set_bnd);
}
//void set_bnd(int b, freal *a) {}

/* apply external density influence */
void add_source(freal *x, freal *s, freal dt)
{
	int i,j;

	#if 0
	//dt=X2I(dt<<10);
	//for (i=SIZE-1; i>=0; --i) x[i]+=(dt*s[i])>>10;
	for (i=SIZE-1; i>=0; --i) x[i]+=s[i]>>6;
	#else
	for (i=SIZE-1; i>=0; --i) x[i]+= XM(dt,s[i]);
	#endif

	#if DEBUG
	printf("add_source()\n");
	for (j=0; j<=HEIGHT+1; j++) {
		printf("s = ");
		for (i=0; i<=WIDTH+1; i++) {
			printf("%04X ", s[IX(i,j)]);
		}
		printf("\n");
	}
	printf("\n");

	for (j=0; j<=HEIGHT+1; j++) {
		printf("x = ");
		for (i=0; i<=WIDTH+1; i++) {
			printf("%04X ", x[IX(i,j)]);
		}
		printf("\n");
	}
	printf("\n");
	#endif
}

void diffuse_bad(int b, freal *x, freal *x0, freal diff, freal dt)
{
	int i,j;
	freal a=XM(dt,diff)*WIDTH*HEIGHT;

	for (i=BORDER; i<WIDTH+BORDER; i++) {
		for (j=BORDER; j<HEIGHT+BORDER; j++) {
			x[IX(i,j)] = x0[IX(i,j)]
			           + XM(a, ( x0[IX(i-1,j)]+x0[IX(i+1,j)]+x0[IX(i,j-1)]+x0[IX(i,j+1)]-4*x0[IX(i,j)] ));
		}
	}
	set_bnd(b, x);
}

/* gauss seidel stabilised.. still grows too much in some cases :/ */
void diffuse(int b, freal *x, freal *x0, freal diff, freal dt)
{
	int i, j, k;
	freal a=XM(dt*WIDTH,diff)*HEIGHT;

	for (k=0; k<20; k++) {
		for (j=BORDER; j<HEIGHT+BORDER; j++) {
			for (i=BORDER; i<WIDTH+BORDER; i++) {
				x[IX(i,j)] = (
			                       X1_0 * x0[IX(i,j)]
				             + a * (x[IX(i-1,j)]+x[IX(i+1,j)]+x[IX(i,j-1)]+x[IX(i,j+1)])
					     )
					     / (X1_0+4*a);
			}
		}
		set_bnd(b, x);
	}
}

void diffuse_fixed(int b, freal *x, freal *x0, freal diff, freal dt)
{
	int i, j, k;
	freal a=XM(dt*WIDTH,diff)*HEIGHT;
	long s1 = ((long long) X1_0<<12)/((X1_0+4*a));
	long s2 = ((long long) a<<12)/((X1_0+4*a));

	for (j=BORDER; j<HEIGHT+BORDER; j++) {
		for (i=BORDER; i<WIDTH+BORDER; ++i) {
			#if FIXEDPOINT
			x[IX(i,j)] =
			(
			  s1*(x0[IX(i,j)]>>4) + s2*((x0[IX(i-1,j)]+x0[IX(i+1,j)]+x0[IX(i,j-1)]+x0[IX(i,j+1)])>>4)
			)>>8;
			#else
			x[IX(i,j)] = (
				       XML( X1_0, x0[IX(i,j)] )
				     + XML( a, (x0[IX(i-1,j)]+x0[IX(i+1,j)]+x0[IX(i,j-1)]+x0[IX(i,j+1)]) )
				     )
				     / (X1_0+4*a);
			#endif
		}
	}
	set_bnd(b, x);

	#if DEBUG
	printf("diffuse_fixed(dt=0x%08X * diff=0x%08X = a=0x%08X)\n", dt, diff, a);
	for (j=0; j<=HEIGHT+1; j++) {
		printf("x0 = ");
		for (i=0; i<=WIDTH+1; i++) {
			printf("%04X ", x0[IX(i,j)]);
		}
		printf("\n");
	}
	printf("\n");

	for (j=0; j<=HEIGHT+1; j++) {
		printf("x = ");
		for (i=0; i<=WIDTH+1; i++) {
			printf("%04X ", x[IX(i,j)]);
		}
		printf("\n");
	}
	printf("\n");
	#endif
}

void advect(int b, freal *d, freal *d0, freal *u, freal *v, freal dt)
{
	int i,j,i0,j0,i1,j1;
	freal x,y,s0,t0,s1,t1,dt0;

	//printf("d=0x%08X d0=0x%08X u=0x%08X v=0x%08X\n", d, d0, u, v);

	dt0=HALVE( (dt * (WIDTH+HEIGHT)) );
	//printf("dt0=0x%X\n", dt0);
	for (j=BORDER; j<HEIGHT+BORDER; j++) {
		//printf("i=%d\n", i);

		for (i=BORDER; i<WIDTH+BORDER; i++) {
			/* backtrack to previous grid position.. */
			//printf("j=%d\t", j);
			x = I2X(i) - XM(dt0, u[IX(i,j)]);
			y = I2X(j) - XM(dt0, v[IX(i,j)]);

			unsigned long xs = (unsigned long) (x-HALVE(X1_0));
			if (xs > I2X(WIDTH)) {
				if ((long) xs < 0)
					x = HALVE(X1_0);
				else if (xs > I2X(WIDTH))
					x=I2X(WIDTH)+HALVE(X1_0);
			}
			i0=X2I(x);
			i1=i0+1;

			#if 1
			unsigned long ys = (unsigned long) (y-HALVE(X1_0));
			if (ys > I2X(WIDTH)) {
				if ((long) ys < 0)
					y = HALVE(X1_0);
				else if (ys > I2X(HEIGHT))
					y=I2X(HEIGHT)+HALVE(X1_0);
			}
			#else
			if (y < HALVE(X1_0))
				y = HALVE(X1_0);
			else if ( y > I2X(HEIGHT)+HALVE(X1_0) )
				y = I2X(HEIGHT)+HALVE(X1_0);
			#endif
			j0=X2I(y);
			j1=j0+1;

			s1=x-I2X(i0);
			s0=X1_0-s1;
			//printf("s0=0x%X ", s0);

			t1=y-I2X(j0);
			t0=X1_0-t1;
			//printf("t0=0x%X ", t0);

			/* bilinear interpolation.. */
			d[IX(i,j)] = XM(s0, (XM(t0, d0[IX(i0,j0)]) + XM(t1, d0[IX(i0,j1)])) )
			           + XM(s1, (XM(t0, d0[IX(i1,j0)]) + XM(t1, d0[IX(i1,j1)])) );
		}
	}
	//set_bnd(b, d);
}

/* bilinear interpolation.. */
#define INTERP() \
i0=X2I(x); \
i1=i0+1; \
j0=X2I(y); \
j1=j0+1; \
s1=(x-I2X(i0))>>11; \
s0=0x100-s1; \
t1=(y-I2X(j0))>>11; \
t0=0x100-t1; \
d[IX(i,j)] = \
( \
  s0 * ((t0 * d0[IX(i0,j0)] + t1 * d0[IX(i0,j1)])>>8 ) \
+ s1 * ((t0 * d0[IX(i1,j0)] + t1 * d0[IX(i1,j1)])>>8 ) \
)>>8;
//i0=X2I(x); \
//i1=i0+1; \
//j0=X2I(y); \
//j1=j0+1; \
//s1=x-I2X(i0); \
//s0=X1_0-s1; \
//t1=y-I2X(j0); \
//t0=X1_0-t1; \
//d[IX(i,j)] = \
//  XM(s0, (XM(t0, d0[IX(i0,j0)]) + XM(t1, d0[IX(i0,j1)])) ) \
//+ XM(s1, (XM(t0, d0[IX(i1,j0)]) + XM(t1, d0[IX(i1,j1)])) );

#define BACKTRACK() \
x = I2X(i) - dt0 * u[IX(i,j)]; \
y = I2X(j) - dt0 * v[IX(i,j)];

/*
 * optimised version with respect to clipping.
 * clipping is done only at the side pixels..
 * under the assumption that a cell cannot move more than one unit per frame..
 * this is reasonable because this is also a criterion for every boundary condition.
 * this is significantly faster..
 */
void advect_new(int b, freal *d, freal *d0, freal *u, freal *v, freal dt)
{
	int i,j,i0,j0,i1,j1;
	freal x,y,s0,t0,s1,t1,dt0;
	//printf("d=0x%08X d0=0x%08X u=0x%08X v=0x%08X\n", d, d0, u, v);
	/* dt0 is big enough and can be respresented as an int.. */
	dt0=X2I(HALVE( dt * (WIDTH+HEIGHT) ));

	/* top case.. check for out of bounds extrapolation.. */
	i=BORDER;
	j=BORDER;
	BACKTRACK();
	if (x < HALVE(X1_0))
		x = HALVE(X1_0);
	if (y < HALVE(X1_0))
		y = HALVE(X1_0);
	INTERP();
	for (i=BORDER+1; i<WIDTH+BORDER-1; i++) {
		/* backtrack to previous grid position.. */
		BACKTRACK();
		if (y < HALVE(X1_0))
			y = HALVE(X1_0);
		INTERP();
	}
	BACKTRACK();
	if (x > I2X(WIDTH)+HALVE(X1_0))
		x=I2X(WIDTH)+HALVE(X1_0);
	if (y < HALVE(X1_0))
		y = HALVE(X1_0);
	INTERP();

	/* body case.. don't check for out of bounds extrapolation in Y dir.. */
	for (j=BORDER+1; j<HEIGHT+BORDER-1; j++) {
		i=BORDER;
		BACKTRACK();
		if (x < HALVE(X1_0))
			x = HALVE(X1_0);
		INTERP();

		for (i=BORDER+1; i<WIDTH+BORDER-1; i++) {
			/* backtrack to previous grid position.. */
			BACKTRACK();
			INTERP();
		}

		BACKTRACK();
		if (x > I2X(WIDTH)+HALVE(X1_0))
			x=I2X(WIDTH)+HALVE(X1_0);
		INTERP();
	}

	/* bottom case.. check for out of bounds extrapolation.. */
	i=BORDER;
	BACKTRACK();
	if (x < HALVE(X1_0))
		x = HALVE(X1_0);
	if (y > I2X(HEIGHT)+HALVE(X1_0))
		y = I2X(HEIGHT)+HALVE(X1_0);
	INTERP();
	for (i=BORDER+1; i<WIDTH+BORDER-1; i++) {
		/* backtrack to previous grid position.. */
		BACKTRACK();
		if (y > I2X(HEIGHT)+HALVE(X1_0))
			y = I2X(HEIGHT)+HALVE(X1_0);
		INTERP();
	}
	BACKTRACK();
	if (x > I2X(WIDTH)+HALVE(X1_0))
		x=I2X(WIDTH)+HALVE(X1_0);
	if (y > I2X(HEIGHT)+HALVE(X1_0))
		y = I2X(HEIGHT)+HALVE(X1_0);
	INTERP();

	//set_bnd(b, d);
}

freal tmpbuf[SIZE];

/* i just don't get the add source in this case.. it only works when x0 (prev_dens) is cleared except for new sources */
void dens_step(freal *x, freal *x0, freal *u, freal *v, freal diff, freal dt)
{
	advect_new(0,x,x0,u,v,dt);
}

#define GS_NUM_LOOPS 20

void project_gauss_seidel(int start_line, int n_lines, freal *div, freal *p)
{
        int i,j,end;

        end = n_lines+start_line;
        for (j = BORDER+start_line; j < end+BORDER; j++) {
                int y_offset=j*(WIDTH+BORDER*2);
                for (i=BORDER; i<WIDTH+BORDER; i++) {
                        p[IX(i,j)] = (
                                        div[IX(i,j)]
                                        + p[IX(i-1,j)]+p[IX(i+1,j)]+p[IX(i,j-1)]+p[IX(i,j+1)]
                                        )/4;
                }
        }
        set_bnd(0,p);
}

freal p[SIZE], div[SIZE];

pthread_t top_worker_thread;
pthread_t btm_worker_thread;
pthread_mutex_t count_mutex;
pthread_cond_t count_threshold_cv;
pthread_mutex_t start_mutex;
pthread_cond_t start_work_cv;
pthread_mutex_t work_done_mutex;
pthread_cond_t work_done_cv;
pthread_mutex_t commandable_mutex;
static int thread_cmd_state;
static int wanna_send, run_threads;
float *map_u, *map_v;
int n_work_iters;
int workers_done, boss_done;

float *top_map_ptr, *btm_map_ptr;

void *work_top(void *idp)
{
        int i, j, top_iters;

        pthread_mutex_lock(&commandable_mutex);
        thread_cmd_state |= 1;
        pthread_mutex_unlock(&commandable_mutex);
        printf("work_top() started\n");

        while(run_threads) {
                /* wait for start signal from main program */
                //    printf("work_top() waiting for work\n");
                pthread_mutex_lock(&start_mutex);
                pthread_cond_wait(&start_work_cv, &start_mutex);
                pthread_mutex_unlock(&start_mutex);
                //printf("work_top() got work\n");

                for (top_iters=0; top_iters<n_work_iters; top_iters++) {
        
                        /* perform the bulk of the work.. */
                        project_gauss_seidel(0, (HEIGHT/2)-1, div, p);
        
                        /* keep the center range of the map locked */
                        pthread_mutex_lock(&count_mutex);
        
                        /* perform work on the center range */
                        project_gauss_seidel((HEIGHT/2)-1, 1, div, p);
        
                        //printf("top: wait for !wanna_send\n");
                        wanna_send=1;
                        for (j=0; wanna_send; j++) {
                                pthread_mutex_unlock(&count_mutex);
                                pthread_yield();
                                pthread_mutex_lock(&count_mutex);
                        } 
                        //printf("top: yielded %d times\n", j);
        
                        /* unlock the center range */
                        pthread_mutex_unlock(&count_mutex);
        
                        //printf("top: iteration %d done.\n", top_iters);
                }

                //printf("top: finished..\n");
		while(!boss_done) {
                pthread_mutex_lock(&work_done_mutex);
                workers_done|=1;
                pthread_cond_signal(&work_done_cv);
                pthread_mutex_unlock(&work_done_mutex);
			pthread_yield();
		}

        }

        printf("top: run_threads==0\n");

        pthread_exit(NULL);
}

void *work_bottom(void *idp)
{
        int i, j, btm_iters;

        pthread_mutex_lock(&commandable_mutex);
        thread_cmd_state |= 2;
        pthread_mutex_unlock(&commandable_mutex);
        printf("work_bottom() started\n");

        while(run_threads) {
                /* wait for start signal from main program */
                //printf("work_bottom() waiting for work\n");
                pthread_mutex_lock(&start_mutex);
                pthread_cond_wait(&start_work_cv, &start_mutex);
                pthread_mutex_unlock(&start_mutex);
                //printf("work_bottom() got work\n");
                //for (j=0; j<n_work_iters; j++) {
                for (btm_iters=0; btm_iters<n_work_iters; btm_iters++) {
        
                        /* wait for top worker to complete its job (the center range should be free) */
                        pthread_mutex_lock(&count_mutex);
                        //printf("work_bottom() wait for signal\n");
                        //printf("btm: wait for wanna_send\n");
                        for (j=0; !wanna_send; j++) {
                                pthread_mutex_unlock(&count_mutex);
                                pthread_yield();
                                pthread_mutex_lock(&count_mutex);
                        } 
                        //printf("btm: yielded %d times\n", j);
        
                        /* perform work using the center range */
                        project_gauss_seidel(HEIGHT/2, 1, div, p);
        
                        wanna_send=0;
        
                        /* unlock the center range */
                        pthread_mutex_unlock(&count_mutex);
        
                        /* perform the bulk of the work.. */
                        project_gauss_seidel((HEIGHT/2)+1, (HEIGHT/2)-1, div, p);
        
                        //printf("btm: iteration %d done.\n", btm_iters);
                }

                /* indicate we've finished the job.. */
                //printf("btm: finished..\n");
		while(!boss_done) {
                pthread_mutex_lock(&work_done_mutex);
                workers_done|=2;
                pthread_cond_signal(&work_done_cv);
                pthread_mutex_unlock(&work_done_mutex);
			pthread_yield();
		}
        }

        printf("btm: run_threads==0\n");

        pthread_exit(NULL);
}

pthread_attr_t attr;

int init_threads()
{
        int ret;

        /* initialise map */
        //memset(map, 0, sizeof(float)*MAP_SIZE);

        /* Initialize mutex and condition variable objects */
        pthread_mutex_init(&count_mutex, NULL);
        pthread_mutex_init(&commandable_mutex, NULL);
        pthread_mutex_init(&start_mutex, NULL);
        pthread_mutex_init(&work_done_mutex, NULL);
        pthread_cond_init (&start_work_cv, NULL);
        pthread_cond_init (&work_done_cv, NULL);

        /*
        For portability, explicitly create threads in a joinable state 
        */
        pthread_attr_init(&attr);
        ret = pthread_attr_setdetachstate(&attr, PTHREAD_CREATE_JOINABLE);
        if (ret)
                return ret;

        /* set threads to default state (alive) */
        run_threads = 1;

        ret = pthread_create(&top_worker_thread, &attr, work_top, (void *) 0);
        if (ret)
                return ret;
        pthread_create(&btm_worker_thread, &attr, work_bottom, (void *) 1);
        if (ret)
                return ret;

        /* wait for worker threads to reach 'commandable' state */
        pthread_mutex_lock(&commandable_mutex);
        while (thread_cmd_state != 3) {
                pthread_mutex_unlock(&commandable_mutex);
                printf("init_threads() yielding\n");
                pthread_yield();
                pthread_mutex_lock(&commandable_mutex);
        }
        pthread_mutex_unlock(&commandable_mutex);

        return 0;
}

/*
 * signal threads to start some serious work
 * precondition: both threads should be active (thread_cmd_state == 3)
 */
int wakeup_workers()
{
        pthread_mutex_lock(&start_mutex);
        pthread_cond_broadcast(&start_work_cv);
        pthread_mutex_unlock(&start_mutex);
}

int provide_work(int n_iters, float *map_u_ptr, float *map_v_ptr)
{
        n_work_iters = n_iters;
        map_u = map_u_ptr;
        map_v = map_v_ptr;
}

int cleanup_threads()
{
        /* lower flag for threads.. */
        run_threads = 0;

        /* wake up threads so that they may quit.. */
        wakeup_workers();

        pthread_join(top_worker_thread, NULL);
        printf("main() top thread joined\n");
        pthread_join(btm_worker_thread, NULL);
        printf("main() btm thread joined\n");

        /* Clean up and exit */
        pthread_attr_destroy(&attr);

        pthread_mutex_destroy(&count_mutex);
        pthread_mutex_destroy(&commandable_mutex);
        pthread_mutex_destroy(&start_mutex);
        pthread_mutex_destroy(&work_done_mutex);
        pthread_cond_destroy(&start_work_cv);
        pthread_cond_destroy(&work_done_cv);

        //pthread_exit (NULL);
}

#define PARALLEL_GS 1
#define BUSY_WAITING_BOSS 1

void project(freal *u, freal *v, int gs_loops)
{
	int i,j,k,m;
	freal h;

	#if 1
	for (i=0; i<WIDTH+BORDER*2; ++i) {
		p[IX(i,0)]=0;
		p[IX(i,HEIGHT+BORDER)]=0;
		div[IX(i,0)]=0;
		div[IX(i,HEIGHT+BORDER)]=0;
	}
	for (j=0; j<HEIGHT+BORDER*2; ++j) {
		p[IX(0,j)]=0;
		p[IX(WIDTH+BORDER,j)]=0;
		div[IX(0,j)]=0;
		div[IX(WIDTH+BORDER,j)]=0;
	}
	#endif

	h = I2X(2)/(HEIGHT+WIDTH);
	long sc = -h/2;
	for (j=BORDER; j<HEIGHT+BORDER; j++) {
		for (i=BORDER; i<WIDTH+BORDER; i++) {
			div[IX(i,j)] = XM( sc, (u[IX(i+1,j)]-u[IX(i-1,j)]+v[IX(i,j+1)]-v[IX(i,j-1)]) );
			p[IX(i,j)]=0;
		}
	}
	set_bnd(0,div);
	set_bnd(0,p);

        /* perform Gauss-Seidel iterations */
        #if PARALLEL_GS

        n_work_iters = gs_loops;
        workers_done = 0;
	boss_done = 0;
        wakeup_workers();
        /* wait for workers to finish (blocked-waiting) */
        //printf("boss: waiting for workers to finish..\n");
        #if BUSY_WAITING_BOSS
	//usleep(10);
	int yields;
        for(yields=0; workers_done != 3; yields++) {
                pthread_yield();
                //usleep(1);
        }
	printf("boss: yielded %d times\n", yields);
        #else
        /* TODO: this is not safe (what if a thread signals before the pthread_cond_wait() is entered?) */
	//timestruct_t to;
	//to.tv_sec = time(NULL) + 2; 
	//to.tv_nsec = 0; 

	while (workers_done != 3) {
		//printf("boss: wait for worker.. (workers_done=%d)\n", workers_done);
	        pthread_mutex_lock(&work_done_mutex);
	        pthread_cond_wait(&work_done_cv, &work_done_mutex);
	        //pthread_cond_wait_timed(&work_done_cv, &work_done_mutex);
	        pthread_mutex_unlock(&work_done_mutex);
		//printf("boss: woke up.. (workers_done=%d)\n", workers_done);
	}
        #endif
	boss_done = 1;
        //printf("boss: good work boys!\n");

        #else
	for (k=0; k<gs_loops; k++) {
                //project_gauss_seidel(0, HEIGHT, div, p);
                project_gauss_seidel(0, HEIGHT/2, div, p);
                project_gauss_seidel(HEIGHT/2, HEIGHT/2, div, p);
	}
        #endif

	freal inv = XDL(X1_0, 2*h);
	long s = (WIDTH+HEIGHT)>>2;
	for (j=BORDER; j<HEIGHT+BORDER; j++) {
		for (i=BORDER; i<WIDTH+BORDER; i++) {
			//u[IX(i,j)] -= XM(p[IX(i+1,j)]-p[IX(i-1,j)], inv);
			//v[IX(i,j)] -= XM(p[IX(i,j+1)]-p[IX(i,j-1)], inv);
			u[IX(i,j)] -= s*(p[IX(i+1,j)]-p[IX(i-1,j)]);
			v[IX(i,j)] -= s*(p[IX(i,j+1)]-p[IX(i,j-1)]);
		}
	}
	set_bnd(1,u);
	set_bnd(2,v);
}

freal tmpu[SIZE], tmpv[SIZE];

void vel_step(freal *u, freal *v, freal visc, freal dt)
{
	#if DEBUG
	printf("vel_step(visc=0x%08X)\n", visc);
	#endif

	diffuse_fixed(1,tmpu,u,visc,dt);
	diffuse_fixed(2,tmpv,v,visc,dt);
	project(tmpu,tmpv,50);
	advect_new(1,u,tmpu,tmpu,tmpv,dt);
	advect_new(2,v,tmpv,tmpu,tmpv,dt);
	project(u,v,50);
}
